from __future__ import absolute_import
import re
from gzip import GzipFile
from random import Random
from io import BytesIO
from StringIO import StringIO


AVERAGE_SPAN_BETWEEN_FLUSHES = 512
APPROX_MIN_FLUSHES = 3
MIN_INTERFLUSH_INTERVAL = 16
FLUSH_LIMIT = 5

class StreamingBuffer(object):
    def __init__(self):
        self.vals = []

    def write(self, val):
        self.vals.append(val)

    def read(self):
        ret = b''.join(self.vals)
        self.vals = []
        return ret

    def flush(self):
        return

    def close(self):
        return

re_accepts_gzip = re.compile(r'\bgzip\b')
cc_delim_re = re.compile(r'\s*,\s*')

def patch_vary_headers(response, newheaders):
    """
    Adds (or updates) the "Vary" header in the given HttpResponse object.
    newheaders is a list of header names that should be in "Vary". Existing
    headers in "Vary" aren't removed.
    """
    # Note that we need to keep the original order intact, because cache
    # implementations may rely on the order of the Vary contents in, say,
    # computing an MD5 hash.
    if response.has_header('Vary'):
        vary_headers = cc_delim_re.split(response['Vary'])
    else:
        vary_headers = []
    # Use .lower() here so we treat headers as case-insensitive.
    existing_headers = set([header.lower() for header in vary_headers])
    additional_headers = [newheader for newheader in newheaders
                          if newheader.lower() not in existing_headers]
    response['Vary'] = ', '.join(vary_headers + additional_headers)


def compress_string(s):

    # avg_block_size is acutally the reciporical of the average
    # intended interflush distance.   

    rnd = Random(s)

    flushes_remaining = FLUSH_LIMIT

    if len(s) < AVERAGE_SPAN_BETWEEN_FLUSHES * APPROX_MIN_FLUSHES:
        avg_block_size = APPROX_MIN_FLUSHES / float(len(s))
    else:
        avg_block_size = 1.0 / AVERAGE_SPAN_BETWEEN_FLUSHES

    s = StringIO(s)
    zbuf = BytesIO()
    zfile = GzipFile(mode='wb', compresslevel=6, fileobj=zbuf)
    chunk = s.read(MIN_INTERFLUSH_INTERVAL + int(rnd.expovariate(avg_block_size)))
    while chunk and flushes_remaining:
        zfile.write(chunk)
        zfile.flush()
        flushes_remaining -= 1
        chunk = s.read(MIN_INTERFLUSH_INTERVAL + int(rnd.expovariate(avg_block_size)))
    zfile.write(chunk)
    zfile.write(s.read())
    zfile.close()
    return zbuf.getvalue()

# Like compress_string, but for iterators of strings.
def compress_sequence(sequence):
    avg_block_size = 1.0 / AVERAGE_SPAN_BETWEEN_FLUSHES

    buf = StreamingBuffer()
    zfile = GzipFile(mode='wb', compresslevel=6, fileobj=buf)
    # Output headers...
    yield buf.read()

    flushes_remaining = FLUSH_LIMIT
    rnd = None
    count = None
    rnd = None
    for item in sequence:
        if rnd is None:
            rnd = Random(0)
            count = int(rnd.expovariate(avg_block_size))
        chunking_buf = BytesIO(item)
        chunk = chunking_buf.read(count)
        while chunk:
            if count is not None:
                count -= len(chunk)
            zfile.write(chunk)
            if count <= 0:
                flushes_remaining -= 1
                zfile.flush()
                yield buf.read()
                if flushes_remaining:
                    count = int(rnd.expovariate(avg_block_size))
                else:
                    count = None
            if count is None:
                chunk = chunking_buf.read()
            else:
                chunk = chunking_buf.read(count)
        zfile.flush()
        yield buf.read()
        if chunk is None:
            break
        
    for item in sequence:
        zfile.write(chunking_buf.read())
        zfile.flush()
        yield buf.read()
        
    zfile.close()
    yield buf.read()

class GZipMiddleware(object):
    """
    This middleware compresses content if the browser allows gzip compression.
    It sets the Vary header accordingly, so that caches will base their storage
    on the Accept-Encoding header.
    """
    def process_response(self, request, response):
        # It's not worth attempting to compress really short responses.

        patch_vary_headers(response, ('Accept-Encoding',))

        # Avoid gzipping if we've already got a content-encoding.
        if response.has_header('Content-Encoding'):
            return response

        # MSIE have issues with gzipped response of various content types.
        if "msie" in request.META.get('HTTP_USER_AGENT', '').lower():
            ctype = response.get('Content-Type', '').lower()
            if not ctype.startswith("text/") or "javascript" in ctype:
                return response

        ae = request.META.get('HTTP_ACCEPT_ENCODING', '')
        if not re_accepts_gzip.search(ae):
            return response

        if getattr(response, 'streaming', False):
            # Delete the `Content-Length` header for streaming content, because
            # we won't know the compressed size until we stream it.
            response.streaming_content = compress_sequence(response.streaming_content)
            del response['Content-Length']
        else:
            # Return the compressed content only if it's actually shorter.
            compressed_content = compress_string(response.content)
            if len(compressed_content) >= len(response.content):
                return response
            response.content = compressed_content
            response['Content-Length'] = str(len(response.content))

        if response.has_header('ETag'):
            response['ETag'] = re.sub('"$', ';gzip"', response['ETag'])
        response['Content-Encoding'] = 'gzip'

        return response



#!/usr/bin/env python

"""
breach_buster
=================

Replacement for Django's gzip middleware.  Protects against BREACH.

"""

from setuptools import setup

setup(
    name='breach_buster',
    version='0.0.2',
    author='Adam DePrince',
    author_email='adeprince@nypublicradio.org',
    description='BREACH resistant gzip middleware for Django',
    url="https://github.com/wnyc/breach_buster",
    long_description=__doc__,
    py_modules=[
        'breach_buster/__init__',
        'breach_buster/middleware/__init__',
        'breach_buster/middleware/gzip',
        'breach_buster/examples/__init__',
        'breach_buster/examples/demo_server',
        ],
    packages=['breach_buster',],
    zip_safe=True,
    license='GPLv3',
    include_package_data=True,
    classifiers=[],
    scripts=[], 
    install_requires=[
        'web.py'
        ]
    )



import web
from StringIO import StringIO
from gzip import GzipFile
from breach_buster.middleware.gzip import compress_string


class breachable:
    def GET(self):
        web.header('content-encoding','gzip')
        out = StringIO()
        f = GzipFile(fileobj=out, mode='w', compresslevel=6)
        name = web.input(name='').name
        f.write(SITE % name)
        f.flush()
        return out.getvalue()

class unbreachable:
    def GET(self):
        web.header('content-encoding','gzip')
        name = web.input(name='').name
        return compress_string(SITE % name)
    
urls = ('/good', 'unbreachable', '/bad', 'breachable')

def main():
    app = web.application(urls, globals())
    app.internalerror = web.debugerror
    app.run()




STORY = """<p>In the olden time, when wishing was having, there lived a King, whose daughters were all beautiful; but the youngest was so exceedingly beautiful that the Sun himself, although he saw her very often, was enchanted every time she came out into the sunshine.</p>

<p>Near the castle of this King was a large and gloomy forest, and in the midst stood an old lime-tree, beneath whose branches splashed a little fountain; so, whenever it was very hot, the King's youngest daughter ran off into this wood, and sat down by the side of this fountain; and, when she felt dull, would often divert herself by throwing a golden ball up in the air and catching it. And this was her favourite amusement.</p>

<a href="do_something?CSRF=valhalla_machine">If you liked this story do something nice for the author</a>


<p>Now, one day it happened, that this golden ball, when the King's daughter threw it into the air, did not fall down into her hand, but on the grass; and then it rolled past her into the fountain. The King's daughter followed the ball with her eyes, but it disappeared beneath the water, which was so deep that no one could see to the bottom. Then she began to lament, and to cry louder and louder; and, as she cried, a voice called out, "Why weepest thou, O King's daughter? thy tears would melt even a stone to pity." And she looked around to the spot whence the voice came, and saw a Frog stretching his thick ugly head out of the water. "Ah! you old water-paddler," said she, "was it you that spoke? I am weeping for my golden ball, which has slipped away from me into the water."</p>

"""
           


SITE = "<html><body><h1>Frog prince</h1><h2>Hello %%s</h2></p><hr/>%s</h></body></html>" % STORY
